theory Hotel
imports Main
begin

abbreviation 
  "SomeFloor" ("(\<lfloor>_\<rfloor>)") where "\<lfloor>x\<rfloor> \<equiv> Some x"

typedecl guest
typedecl key
types card = "key \<times> key"
typedecl room

record state =
(* reception *)
owns   ::   "room \<Rightarrow> guest option"
currk  ::   "room \<Rightarrow> key"
issued ::   "key set"
(* guests *)
cards  ::   "guest \<Rightarrow> card set"
(* rooms *)
roomk  ::   "room \<Rightarrow> key"
isin   ::   "room \<Rightarrow> guest set"
(* ghost variable *)
safe   ::   "room \<Rightarrow> bool"

inductive_set reach :: "state set"
where
init: (* initk injective *)
  "\<forall>r r'. (initk r = initk r') = (r = r') \<Longrightarrow>
  \<lparr> owns = (\<lambda>r . None), currk = initk, issued = range initk,
    cards = (\<lambda>g. {}), roomk = initk, isin = (\<lambda>r. {}),
    safe = (\<lambda>r. True) \<rparr> \<in> reach"
| check_in:
  "\<lbrakk> s \<in> reach; k \<notin> issued s \<rbrakk> \<Longrightarrow>
  s\<lparr> currk := (currk s)(r := k), 
     issued := issued s \<union> {k},
     cards := (cards s)(g := cards s g \<union> {(currk s r,k)}),
     owns := (owns s)(r := \<lfloor>g\<rfloor>),
     safe := (safe s)(r := False) \<rparr> \<in> reach"
| enter_room:
  "\<lbrakk> s \<in> reach; (k,k') \<in> cards s g; roomk s r \<in> {k,k'} \<rbrakk> \<Longrightarrow>
  s\<lparr> isin := (isin s)(r := isin s r \<union> {g}),
     roomk := (roomk s)(r := k'),
     safe := (safe s)(r := owns s r = \<lfloor>g\<rfloor> \<and> isin s r = {} \<and> k' = currk s r 
                           \<or> safe s r)
   \<rparr> \<in> reach"
| exit_room:
  "\<lbrakk> s \<in> reach; g \<in> isin s r\<rbrakk> \<Longrightarrow>
  s\<lparr> isin := (isin s)(r := isin s r - {g}) \<rparr> \<in> reach"

(* invariant 1 *)
lemma currk_issued[simp]: "s \<in> reach \<Longrightarrow> currk s r \<in> issued s"
by (induct set: reach) auto

lemma key2_issued[simp]: "\<lbrakk> s \<in> reach; (k,k') \<in> cards s g \<rbrakk> \<Longrightarrow> k' \<in> issued s"
by (induct set: reach) auto

lemma key1_issued[simp]: "\<lbrakk> s \<in> reach; (k,k') \<in> cards s g \<rbrakk> \<Longrightarrow> k \<in> issued s"
by (induct set: reach) auto

lemma roomk_issued[simp]: "s \<in> reach \<Longrightarrow> roomk s k \<in> issued s"
by (induct set: reach) auto

lemma currk_inj[simp]:
 "s \<in> reach \<Longrightarrow> ∀r r'. (currk s r = currk s r') = (r = r')"
by (induct set: reach) auto

lemma key1_not_currk[simp]:
 "s \<in> reach \<Longrightarrow> (currk s r,k') \<notin> cards s g"
by (induct set: reach) auto

lemma guest_key2_disj2[simp]:
 "\<lbrakk> s \<in> reach; (k1,k) ∈ cards s g1; (k2,k) ∈ cards s g2 \<rbrakk> \<Longrightarrow> g1=g2"
by (induct set: reach) (auto simp:key2_not_currk)

lemma safe_roomk_currk[simp]:
 "s \<in> reach; safe s r\<rbrakk> \<Longrightarrow> roomk s r = currk s r"
by (induct set: reach) auto

lemma only_owner_enter_normal[simp]:
 "\<lbrakk> s \<in> reach; safe s r; (k',roomk s r) ∈ cards s g \<rbrakk> \<Longrightarrow> owns s r = \<lfloor>g\<rfloor>"
by (induct set: reach) auto

theorem "\<lbrakk>s : reach; safe s r; g \<in> isin s r\<rbrakk> \<Longrightarrow> owns s r = \<lfloor>g\<rfloor>"


